/*
 * Yggdrasil AGI :: Copyright (C) 1998 Dark Fiber
 */


#include "includes.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
Determine what AGI v2 system to emulate, these are the "majour" version
to emulate, thus 2.915 comes under 2.917, 2.4xxx is 2.440, etc.

0x2089
0x2272
0x2440
0x2917
0x2936
*/


typedef struct v2_gID
{
	UINT16	id;
	UINT32	crc;
} v2_gID;


UINT32 v2id_game(void)
{
	v2_gID	v2id[]=
		{
            {0x2089, 0x68036},      /* Space Quest 1,			2.089 */

            {0x2272, 0x31677},      /* XMAS Card 1986,          2.272 */
            {0x2272, 0x67F6E},      /* Space Quest 1,           2.272 */
            {0x2272, 0x88673},      /* Kings Quest 3,           2.272 */

            {0x2440, 0x24A18},      /* Black Cauldron           2.439 */
            {0x2440, 0x633CB},      /* Kings Quest 2,           2.411 */
            {0x2440, 0x68244},      /* Space Quest 1,           2.426 */
            {0x2440, 0x68244},      /* Space Quest 1,           2.917 */
            {0x2440, 0x6F5E1},      /* Larry 1,                 2.440 */
            {0x2440, 0x6F5E1},      /* Larry 1,                 2.917 */

            {0x2917, 0x49EDA},      /* Kings Quest 1,           2.917 */
            {0x2917, 0x5D77C},      /* Mixed Up Mother Goose,   2.915 */
            {0x2917, 0x63338},      /* Kings Quest 2,           2.917 */
            {0x2917, 0x7EF06},      /* Police Quest 1,          2.917 */
            {0x2917, 0x843AC},      /* Demo 3,					2.917 */
            {0x2917, 0x8856C},      /* Demo 2,					2.915 */
            {0x2917, 0x8CC43},      /* Demo 1,					2.915 */
            {0x2917, 0x8DA3E},      /* Space Quest 2,			2.915 */

            {0x2936, 0x8410B},      /* Kings Quest 3,			2.936 */
            {0x2936, 0x8E29B},      /* Space Quest 2,			2.936 */

			{0x0,	0x0}
		};
	UINT8	*fn[]=
		{
			(UINT8*)"viewdir",
			(UINT8*)"logdir",
			(UINT8*)"picdir",
			(UINT8*)"snddir",
			(UINT8*)"words.tok",
			(UINT8*)"object",
			(UINT8*)""
		};
	UINT16	ec=err_OK, x, y;
	UINT32	len, c, crc;
	UINT8	*buff;
	FILE	*fp;

	loader->int_version=0x2917;
	buff=(UINT8*)malloc(8192);

	for(crc=0, y=0; fn[y][0]!=0x0; y++)
	{
		fixpath(NO_GAMEDIR, fn[y]);
		if((fp=fopen((char*)path, "rb"))!=NULL)
		{
			len=1;
			while(len>0)
			{
				memset(buff, 0x0, 8192);
				len=fread(buff, 1, 8000, fp);
				for(c=0; c<len; c++)
					crc+=*(buff+c);
			}
			fclose(fp);
		}
	}
	free(buff);

	for(x=0; v2id[x].id!=0x0; x++)
	{
		if(v2id[x].crc==crc)
		{
			loader->int_version=v2id[x].id;
			/*
			printf("Detected a Sierra game, setting known version\n");
			*/
		}
	}

	/* setup the differences in the opcodes and other bits in the
	   AGI v2 specs
	*/

	if(optEmuVersion!=0x0)
		loader->int_version=optEmuVersion;

	switch(loader->int_version)
	{
		case 0x2089:
			logic_names_cmd[0x86].num_args=0;	/* quit : 0 args */
		case 0x2272:
			logic_names_cmd[0x97].num_args=3;	/* print.at : 3 args */
			logic_names_cmd[0x98].num_args=3;	/* print.at.v : 3 args */
			break;
		case 0x2440:
			break;
		case 0x2917:
			break;
		case 0x2936:
			break;
		default:
			printf("** Cannot setup for unknown version\n");
			ec=err_UnknownAGIVersion;
			break;
	}

	return ec;
}

/*
Currently, there is no known difference between v3.002.098 -> v3.002.149
So version emulated;

0x0086,
0x0149
*/

UINT32 v3id_game(void)
{
	v2_gID	v3id[]=
		{
            {0x3086, 0x0B124B},     /* Kings Quest 4,       3.002.086 */
            {0x3149, 0x023E0E},     /* Black Cauldron,      3.002.098 */
            {0x3149, 0x089592},     /* Demo 4,              3.002.102 */
            {0x3149, 0x09CB15},     /* Kings Quest 4 Demo,  3.002.102 */
            {0x3149, 0x04C16D},     /* Manhunter 1,         3.002.107 */
            {0x3149, 0x053971},     /* Manhunter 2,         3.002.149 */
            {0x3149, 0x0B1C08},     /* Gold Rush,           3.002.149 */

			{0x0, 0x0}
		};
	UINT8	*fn[]=
		{
			(UINT8*)"words.tok",
			(UINT8*)"object",
			(UINT8*)""
		};

	UINT16	ec=err_OK, x, y;
	UINT32	len, c, crc;
	UINT8	*buff;
	FILE	*fp;

	loader->int_version=0x3149;			/* default v3 emulation */
	buff=(UINT8*)malloc(8192);

	for(crc=0, y=0; fn[y][0]!=0x0; y++)
	{
		fixpath(NO_GAMEDIR, fn[y]);
		if((fp=fopen((char*)path, "rb"))!=NULL)
		{
			len=1;
			while(len>0)
			{
				memset(buff, 0x0, 8192);
				len=fread(buff, 1, 8000, fp);
				for(c=0; c<len; c++)
					crc+=*(buff+c);
			}
			fclose(fp);
		}
	}

	/* no do the directory file */

	fixpath(GAMEDIR, (UINT8*)DIR);
	if((fp=fopen((char*)path, "rb"))!=NULL)
	{
		len=1;
		while(len>0)
		{
			memset(buff, 0x0, 8192);
			len=fread(buff, 1, 8000, fp);
			for(c=0; c<len; c++)
				crc+=*(buff+c);
		}
		fclose(fp);
	}

	free(buff);

	for(x=0; v3id[x].id!=0x0; x++)
	{
		if(v3id[x].crc==crc)
		{
			loader->int_version=v3id[x].id;
			/*
			printf("Detected a Sierra game, setting known version\n");
			*/
		}
	}


	if(optEmuVersion!=0x0)
		loader->int_version=optEmuVersion;

	switch(loader->int_version)
	{
		case 0x3086:
			logic_names_cmd[0xAD].num_args=1;	/* 173 : 1 args */
			break;
		case 0x3149:
			logic_names_cmd[0xAD].num_args=0;	/* 173 : 0 args */
			break;
		default:
			printf("** Cannot setup for unknown version\n");
			ec=err_UnknownAGIVersion;
			break;
	}
	return ec;
}
