/*
 * Yggdrasil AGI :: Copyright (C) 1998 Dark Fiber
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>

#include "includes.h"

UINT16	run_game(void);
UINT16	run_game2(void);
UINT16	detect_game(UINT8* fname);

UINT16	show_words(void);
UINT16	show_objects(void);
UINT16	view_pictures(void);

void	interpret_cycle(void);
void	update_objects(void);
void	calc_obj_motion(void);
void	normal_motion(UINT8 entry, SINT8 x, SINT8 y);
void	adj_pos(UINT8 entry, UINT8 x, UINT8 y);



int main(int argc, char *argv[])
{
	UINT16	ec;

	printf("" TITLE " v" VERSION "; Release " RELEASE ", by " AUTHOR "\n"
			"A Sierra On-Line(tm) Adventure Game Interpreter.\n\n");

	ec=parse_cli((UINT16*)&argc, (UINT8**)argv);
	if(ec==err_OK)
		ec=init_machine();

	if(ec==err_OK)
	{
		loader=(AGI_LOADER*)NULL;

		if(argc>1)
			ec=detect_game((UINT8*)argv[1]);
		else
			ec=detect_game((UINT8*)".");

		if(ec==err_OK)
		{
			do
			{
    			printf("AGI v%i game detected.\n", loader->version);
    			/*printf("AGI loader v%s\n", loader->ver_string);*/

    			ec=agi_init();
    			if(ec==err_OK)
    			{
    				/* setup machine specif AGI flags, etc */
    				ec=setup_machine();

    				horizon=36;
    				control_mode=program_control;

    				o_status=5;

    				ec=run_game();
    			}
    			/* deinit our resources */
    			agi_deinit();
			}while(ec==err_RestartGame);
		}
		else
			ec=err_InvalidAGIFile;
	}

	if(ec!=err_OK && ec!=err_DoNothing)
	{
		printf("Error : %04i : ", ec);
		switch(ec)
		{
			case err_BadCLISwitch: printf("Bad CLI switch.\n"); break;
			case err_InvalidAGIFile: printf("Invalid AGI file, or no AGI file in current directory.\n"); break;
			case err_BadFileOpen: printf("Unable to open file.\n"); break;
			case err_NotEnoughMemory: printf("Not enough memory.\n"); break;
			case err_BadResource: printf("Error in resource.\n"); break;
			case err_UnknownAGIVersion: printf("Unknown AGI version.\n"); break;
		}
		printf("\nFor extended help, use -? for syntax\n");
	}

	if(gname!=NULL)
		free(gname);

	deinit_machine();

	printf("\n");
	exit(ec);			/* this is for the memory system */
	return ec;
}


UINT16 detect_game(UINT8 *gn)
{
	UINT16	ec=err_OK;

	if(gname==NULL)
		gname=(UINT8*)string_dupe((char*)null_msg, __LINE__, __FILE__);

#ifdef __AGI_V2
	loader=&agi_v2;
	ec=loader->detect_game(gn);
#endif

#ifdef __AGI_V3
	if(ec!=err_OK)
	{
		loader=&agi_v3;
		ec=loader->detect_game(gn);
	}
#endif

	return(ec);
}

UINT16 show_words(void)
{
	UINT16 ec=err_OK;
	UINT16 count;
	UINT16	uid, sid, lid;

	uid=0;
	sid=0;
	lid=0xFFFF;

	printf("  ID   Word\n");
	for(count=0; count<num_words; count++)
	{
		if(lid==words[count].id)
			sid++;
		else
		{
			lid=words[count].id;
			uid++;
		}
		printf("%4i - %s\n", words[count].id, words[count].word);
	}

	printf("\n%li words in dictionary\n", num_words);
	printf("%5i unique words\n", uid);
	printf("%5i synonyms\n", sid);
	printf("\nThis is dodgy, only sysnonyms are id'd if they follow each other\n"
    	"eg, space quest 1,  knife, army knife, xenon army knife are all synonyms\n"
    	"but are not treated as such as they do not alphabetically follow each other.\n");

	return ec;
}

UINT16 show_objects(void)
{
	UINT16 ec=err_OK;
	UINT16 count;

	printf(" ID   Objects\n");
	for(count=0; count<num_objects; count++)
		printf("%2i - %s\n", (objects+count)->location, (objects+count)->name);

	printf("\n%li objects\n", num_objects);
	return ec;
}

UINT16 run_game(void)
{
	UINT16	ec=err_OK;

	switch(optGameRun)
	{
		case gRUN_GAME:
			ec=run_game2();
			break;

		case gVIEW_PICTURES:
			ec=view_pictures();
			break;

		case gSHOW_WORDS:
			ec=show_words();
			break;

		case gSHOW_OBJECTS:
			ec=show_objects();
			break;
	}
	return ec;
}


UINT16 view_pictures(void)
{
	UINT16	ec=err_OK;
	UINT8	flag=0;
	UINT32	resnum=0;
	UINT8	x[64];
	SINT16	dir=1;
	SINT16	xx;

	gfx->init_video_mode();
	gfx->set_palette(palette, 0, 16);

	while(flag==0 && ec==err_OK)
	{
		xx=resnum;
		x[0]=0;
		x[1]=0;
		while(x[1]==0)
		{
			/* scan for resource */
			if(dir_pic[xx].offset==EMPTY)
			{
				xx+=dir;
				if(xx<0)
					xx=MAX_DIRS-1;
				if(xx>MAX_DIRS-1)
					xx=0;
				x[0]++;
				if(x[0]==0)
					x[1]=2;
			}
			else
				x[1]=1;
		}
		resnum=xx;

		if(x[1]!=2)
		{
    		ec=loader->load_resource(rPICTURE, resnum);
    		if(ec==err_OK)
    		{
    		    /* decodes the raw data to useable form */
    			decode_picture(resnum);
    			dump_screenX();

            	sprintf((char*)x, "V:Vis C:Con P:Pri Q:Quit +:Next -:Prev");
            	gfx->set_cursor_pos(23, 1);
            	printf("%s", x);
            	fflush(stdout);
    			sprintf((char*)x, "Picture %3li", resnum);
    			gfx->set_cursor_pos(24, 1);
				printf("%s", x);
				fflush(stdout);

    			x[0]=1;
    			while(x[0]!=0)
    			{

    				x[0]=tolower(getch());
    				switch(x[0])
    				{
    					case 'q':
    						flag=1;
    						x[0]=0;
    						break;
    					case 'v':
    						dump_screenX();
    						break;
    					case 'p':
    						dump_pri(resnum);
    						break;
    					case 'c':
    						dump_con(resnum);
    						break;
    					case 'x':
    						dump_x(resnum);
    						break;
    					case '+':
    						xx=resnum;
    						if(xx<MAX_DIRS-1)
    							xx++;
    						else
    							xx=0;
    						x[0]=0;
    						dir=1;
    						break;
    					case '-':
    						xx=resnum;
    						if(xx>0)
    							xx--;
    						else
    							xx=MAX_DIRS-1;
    						x[0]=0;
    						dir=-1;
    						break;
    				}
    			}
    			loader->unload_resource(rPICTURE, resnum);
    			resnum=xx;
    		}
    	}
    	else
    		ec=1;		/* no pics? */
	}
	gfx->deinit_video_mode();
	return ec;
}

UINT16 run_game2(void)
{
	UINT16	ec=err_OK;
	UINT32	x;
	UINT16	counter;

	gfx->init_video_mode();
	gfx->set_palette(palette, 0, 16);

	cmd_version();
	setflag(F_logic_zeron_firsttime, __TRUE);
	setflag(F_new_room_exec, __TRUE);
	setflag(F_restart_game, __FALSE);

	allow_kyb_input=__TRUE;
	counter=0;
	new_room_num=0;
	quit_prog_now=__FALSE;

	do
	{
#ifndef __DJGPP__
		for(x=0xFFFFF; x>0; x--);
#else
		for(x=0x3FFFFF; x>0; x--);
#endif
		counter++;

		if(counter==20)
		{
			counter=0;
			setvar(V_seconds, getvar(V_seconds)+1);
			if(getvar(V_seconds)==60)
			{
				setvar(V_seconds, 0);
				setvar(V_minutes, getvar(V_minutes)+1);
				if(getvar(V_minutes)==60)
				{
					setvar(V_minutes, 0);
					setvar(V_hours, getvar(V_hours)+1);
					if(getvar(V_hours)==24)
					{
						setvar(V_hours, 0);
						setvar(V_days, getvar(V_days)+1);
					}
				}
			}
		}

		if(counter%(1+getvar(V_time_delay))==0)
		{
			interpret_cycle();
			gfx->put_screen();
		}

		if(quit_prog_now==0xFF)
		{
			ec=err_RestartGame;
			quit_prog_now=__TRUE;
		}
		if(scan_code==0x4F)
		{
			quit_prog_now=__TRUE;
			kill_sprites();
		}
		if(scan_code==0x47)
			optDebug=__TRUE;
	}while(quit_prog_now==__FALSE);

	gfx->deinit_video_mode();
	return ec;
}


void interpret_cycle(void)
{
	/* run a cycle */

	setflag(F_entered_cli, __FALSE);
	setflag(F_said_accepted_input, __FALSE);
	setvar(V_key, 0x0);

	poll_keyboard();
    view_table[0].direction=getvar(6);
	calc_obj_motion();
	update_status_line(__FALSE);

	do
	{
		ego_in_new_room=__FALSE;
		exit_all_logics=__FALSE;

		/* set current start IP */
		logics[0].cIP=logics[0].sIP;
		run_logic(0);

		/* quit built in debugger command */
		if(optDebug==3)
			optDebug=__TRUE;


		/* make sure logic 0 is not set to 'firsttime' */
		setflag(F_logic_zeron_firsttime, __FALSE);

		if(quit_prog_now==__FALSE)
		{
    		view_table[0].direction=getvar(6);
			update_status_line(__FALSE);

    		setvar(5, 0);
    		setvar(4, 0);
    		setflag(5, __FALSE);
    		setflag(6, __FALSE);
    		setflag(12, __FALSE);

			if(ego_in_new_room==__FALSE)
				update_objects();
			else
    			new_room(new_room_num);
    	}
	}while(ego_in_new_room==__TRUE && quit_prog_now==__FALSE && exit_all_logics==__FALSE);
}

void update_objects(void)
{
	UINT16	count1;
	SINT16	ccel;

	release_sprites();

	/* now copy the bitmaps onto the screen */
	for(count1=0; count1<MAX_VIEWTABLE; count1++)
	{
		if(((view_table[count1].flags&ANIMATED)==ANIMATED) && ((view_table[count1].flags&DRAWN)==DRAWN))
		{
			if((view_table[count1].flags&UPDATE)==UPDATE && (view_table[count1].flags&CYCLING)==CYCLING)
			{
				view_table[count1].cycle_time_count++;
				if(view_table[count1].cycle_time_count > view_table[count1].cycle_time)
				{
					view_table[count1].cycle_time_count=1;
					ccel=view_table[count1].cur_cel;
					switch(view_table[count1].cycle_status)
					{
						case 0:				/* normale cycle status */
							ccel++;
							if(ccel>=view_table[count1].num_cels)
								ccel=0;
							set_cel(count1, ccel);
							break;
						case 1:				/* end of loop */
							ccel++;
							if(ccel>=view_table[count1].num_cels)
							{
								view_table[count1].flags&=~CYCLING;
								setflag(view_table[count1].parm1, __TRUE);
							}
							else
								set_cel(count1, ccel);
							break;
						case 2:				/* reverse loop */
							ccel--;
							if(ccel<0)
							{
								setflag(view_table[count1].parm1, __TRUE);
								view_table[count1].flags&=~CYCLING;
							}
							else
								set_cel(count1, ccel);
							break;
						case 3:				/* reverse cycle */
							ccel--;
							if(ccel<0)
								ccel=view_table[count1].num_cels-1;
							set_cel(count1, ccel);
							break;
					}
				}
			}

			/* save co-ords for putting the images back */
			view_table[count1].bg_x=view_table[count1].x_pos;
			view_table[count1].bg_y=view_table[count1].y_pos-view_table[count1].y_size;

			/* copy the pic into bg_scr */
			view_table[count1].bg_scr=(UINT8*)malloc(view_table[count1].x_size * view_table[count1].y_size);
			get_bitmap(view_table[count1].bg_scr, screen_data, view_table[count1].bg_x, view_table[count1].bg_y, view_table[count1].x_size, view_table[count1].y_size);

			view_table[count1].bg_pri=(UINT8*)malloc(view_table[count1].x_size * view_table[count1].y_size);
			get_bitmap(view_table[count1].bg_pri, xdata_data, view_table[count1].bg_x, view_table[count1].bg_y, view_table[count1].x_size, view_table[count1].y_size);

			if(!((view_table[count1].flags&FIXED_PRIORITY)==FIXED_PRIORITY))
			{
				if(view_table[count1].y_pos<60)
					view_table[count1].priority=4;
				else
					view_table[count1].priority=(view_table[count1].y_pos/12)+1;
			}
		}
	}

	/* draw all cels */
	for(count1=0; count1<MAX_VIEWTABLE; count1++)
	{
		if((view_table[count1].flags&ANIMATED)==ANIMATED && (view_table[count1].flags&DRAWN)==DRAWN)
		{
			agi_put_bitmap(view_table[count1].cel->data,
				view_table[count1].x_pos,
				view_table[count1].y_pos-view_table[count1].y_size,
				view_table[count1].x_size, view_table[count1].y_size,
				view_table[count1].cel->transparency&0xF, view_table[count1].priority);
		}
	}
}

void calc_obj_motion(void)
{
	UINT8	em;
	UINT8	ox, oy;

	for(em=0; em<MAX_VIEWTABLE; em++)
	{
		if((view_table[em].flags&MOTION)==MOTION && (view_table[em].flags&UPDATE)==UPDATE)
		{
			view_table[em].step_time_count+=view_table[em].step_size;
			if(view_table[em].step_time_count > view_table[em].step_time)
			{
				view_table[em].step_time_count=1;
				switch(view_table[em].motion)
				{
					case 0:				/* normal */
						switch(view_table[em].direction)
						{
							case 0: break;		/* middle */
							case 1: normal_motion(em, 0, -1); break;
							case 2: normal_motion(em, 1, -1); break;
							case 3: normal_motion(em, 1, 0); break;
							case 4: normal_motion(em, 1, 1); break;
							case 5: normal_motion(em, 0, 1); break;
							case 6: normal_motion(em, -1, 1); break;
							case 7: normal_motion(em, -1, 0); break;
							case 8: normal_motion(em, -1, -1); break;
						}
						break;
					case 1:				/* wander */
						ox=view_table[em].x_pos;
						oy=view_table[em].y_pos;
						switch(view_table[em].direction)
						{
							case 0: break;		/* middle */
							case 1: normal_motion(em, 0, -1); break;
							case 2: normal_motion(em, 1, -1); break;
							case 3: normal_motion(em, 1, 0); break;
							case 4: normal_motion(em, 1, 1); break;
							case 5: normal_motion(em, 0, 1); break;
							case 6: normal_motion(em, -1, 1); break;
							case 7: normal_motion(em, -1, 0); break;
							case 8: normal_motion(em, -1, -1); break;
						}
						if(view_table[em].x_pos==ox && view_table[em].y_pos==oy)
							view_table[em].direction=rnd(8)+1;
						break;
					case 2:					/* follow ego */
						break;
					case 3:					/* move obj */
						adj_pos(em, view_table[em].parm1, view_table[em].parm2);

						if(em==0)
							setvar(V_ego_dir, 0);

						if(view_table[em].x_pos==view_table[em].parm1 && view_table[em].y_pos==view_table[em].parm2)
						{
							setflag(view_table[em].parm4, __TRUE);
							view_table[em].step_size=view_table[em].parm3;
							view_table[em].flags&=~MOTION;
							view_table[em].motion=0;
						}
						break;
				}
			}
		}
		calc_direction(em);
	}
}

void normal_motion(UINT8 em, SINT8 x, SINT8 y)
{
	SINT16	tmx, tmy, tsx, sx, ex, wc;
	UINT8	flag=0, z;

	tmx=view_table[em].x_pos+x;
	tmy=view_table[em].y_pos+y;

	if(em==0)
	{
		if(tmx<0)
		{
			setvar(V_border_touch_ego, 4);
			return;
		}
		else
		{
			if(tmx>(160-view_table[em].x_size))
			{
				setvar(V_border_touch_ego, 2);
				return;
			}
			else
			{
				if(tmy>167)
				{
					setvar(V_border_touch_ego, 3);
					return;
				}
				else
				{
					if(tmy<horizon)
					{
						setvar(V_border_touch_ego, 1);
						return;
					}
				}
			}
		}
	}

	if(em==0)
	{
		setflag(F_ego_water, __FALSE);
		setflag(F_ego_touched_p2, __FALSE);
		sx=tmx;
		ex=sx+view_table[em].x_size;

		/* do control lines n shit in here */
		for(wc=0, tsx=sx; tsx<ex; tsx++)
		{
			switch(control_data[tmy*160+tsx])
			{
				case 0:			/* unconditional black. no go at all! */
					return;
				case 1:			/* conditional blue */
					if((view_table[em].flags&IGNORE_BLOCKS)!=0)
						return;
					break;
				case 2:			/* trigger */
					setflag(3, __TRUE);
					view_table[em].x_pos=tmx;
					view_table[em].y_pos=tmy;
					return;
				case 3:			/* water? */
					wc++;
					break;
			}
		}
		if(flag==0 && wc==view_table[em].x_size)
		{
			view_table[em].x_pos=tmx;
			view_table[em].y_pos=tmy;
			setflag(0, __TRUE);
			return;
		}
	}
	else
	{
		sx=tmx;
		ex=sx + view_table[em].x_size;
		for(tsx=sx; tsx<ex; tsx++)
		{
			if(tmy<horizon && tmy>=167)
				return;
			if(tsx<=0 && tsx>=160)
				return;

			///*
			z=xdata_data[tmy*160 + tsx];
			if((view_table[em].flags&ON_WATER)==ON_WATER && z!=3)
				return;
			if((view_table[em].flags&ON_LAND)==ON_LAND && z!=4)
				return;
			//*/

			/*
			if((view_table[em].flags&ON_WATER)==ON_WATER && control_data[tmy*160+tsx]!=3)
				return;
			if((view_table[em].flags&ON_LAND)==ON_LAND && control_data[tmy*160+tsx]!=4)
				return;
			*/
		}
	}

	view_table[em].x_pos=tmx;
	view_table[em].y_pos=tmy;
}

void adj_pos(UINT8 em, UINT8 x, UINT8 y)
{
	UINT8	x2, y2;
	UINT8	z;

	/* crappy routine */
	x2=view_table[em].x_pos;
	y2=view_table[em].y_pos;
	z=view_table[em].step_size;

	if(x2<x)
	{
		x2+=z;
		if(x2>x)
			x2=x;
	}
	else if(x2>x)
	{
		x2-=z;
		if(x2<x)
			x2=x;
	}

	if(y2<y)
	{
		y2+=z;
		if(y2>y)
			y2=y;
	}
	else if(y2>y)
	{
		y2-=z;
		if(y2<y)
			y2=y;
	}

/*
	if(y2<8)
		y=8;
	if(x2+view_table[em].x_size>160)
		x2=160;
	if(y2>168)
		y2=167;
*/

	view_table[em].x_pos=x2;
	view_table[em].y_pos=y2;
}
